""" Objects as stored in the Tql database.
"""

from fnmatch import fnmatch

from cloudcontrol.common.tql.db.tag import StaticTag


class TqlObject(object):

    """ An object in the TQL database.
    """

    def __init__(self, id_):
        id_ = StaticTag('id', id_)
        self._tags = {'id': id_}

    def __repr__(self):
        return '<TqlObject id=%s>' % self.id

    def __contains__(self, value):
        return value in self._tags

    def __hash__(self):
        return hash(self._tags['id'].value)

    def __cmp__(self, other):
        return cmp(self.id, other.id)

    def __getitem__(self, name):
        return self._tags[name]

    def __iter__(self):
        return iter(self.itertags())

    @property
    def id(self):
        """ Get the id of the object.
        """
        return self._tags['id'].value

    def itertags(self):
        """ Iter over tags registered on this object.
        """
        for tag in self._tags.itervalues():
            yield tag

    def itermatchingtags(self, patterns):
        """ Iter over tags matching given patterns.
        """
        if isinstance(patterns, basestring):
            patterns = (patterns,)

        for tag in self._tags.itervalues():
            for pattern in patterns:
                if fnmatch(tag.name, pattern):
                    yield tag
                    break

    def get(self, name, default=None):
        """ Get the tag with the specified name (or default).
        """
        return self._tags.get(name, default)

    def get_value(self, name, default=None):
        """ Get the value of the tag with the specified name (or default).
        """
        tag = self.get(name, None)
        return tag if tag is not None else default

    def register(self, tag):
        """ Register a tag on this object.
        """

        if tag.name in self._tags:
            raise KeyError('A tag with this name is '
                           'already registered on this object')
        else:
            self._tags[tag.name] = tag

    def unregister(self, name):
        """ Unregister the tag with specified name.
        """
        del self._tags[name]
